from ..client import ResponseStructureError
from .contact import Contact
from .subscription import Subscription


class UserSettings:
    def __init__(self, login, contacts, subscriptions):
        self.login = login
        self.contacts = contacts
        self.subscriptions = subscriptions


class UserManager:
    def __init__(self, client):
        self._client = client

    def get_username(self):
        """
        Gets the username of the authenticated user if it is available.

        :return: login

        :raises: ResponseStructureError
        """
        result = self._client.get(self._full_path())
        if 'login' not in result:
            raise ResponseStructureError("'login' field doesn't exist in response", result)
        return result['login']

    def get_user_settings(self):
        """
        Get the user's contacts and subscriptions.

        :return: user settings

        :raises: ResponseStructureError
        """
        result = self._client.get(self._full_path('settings'))
        required = ['login', 'contacts', 'subscriptions']
        for field in required:
            if field not in result:
                raise ResponseStructureError("'{}' field doesn't exist in response".format(field), result)

        contacts = []
        for contact in result['contacts']:
            contacts.append(Contact(**contact))
        result['contacts'] = contacts

        subscriptions = []
        for subscription in result['subscriptions']:
            subscriptions.append(Subscription(self._client, **subscription))
        result['subscriptions'] = subscriptions

        return UserSettings(**result)

    def _full_path(self, path=''):
        if path:
            return 'user/{}'.format(path)
        return 'user'



from ..client import InvalidJSONError
from ..client import ResponseStructureError


class NotificationManager:
    def __init__(self, client):
        self._client = client

    def fetch_all(self):
        """
        Returns all notifications
        :return: list of dict

        :raises: ResponseStructureError
        """
        result = self.fetch(start=0, end=-1)
        return result

    def fetch(self, start, end):
        """
        Gets a paginated list of notifications
        :return: list of dict

        :param start
        :param end

        :raises: ResponseStructureError
        """
        params = {
            'start': start,
            'end': end
        }
        result = self._client.get(self._full_path(), params=params)
        if 'list' not in result:
            raise ResponseStructureError("list doesn't exist in response", result)

        return result['list']

    def delete_all(self):
        """
        Remove all notifications

        :return: True on success, False otherwise
        """
        try:
            result = self._client.delete(self._full_path("all"))
            return False
        except InvalidJSONError as e:
            if e.content == b'':  # successfully if response is blank
                return True
            return False

    def delete(self, notification_id):
        """
        Remove notification by id

        :param notification_id: str notification id

        :return: True on success, False otherwise
        """

        params = {
            'id': notification_id,
        }
        try:
            result = self._client.delete(self._full_path(), params=params)
            if 'result' in result and result['result'] == 0:
                return True
        except InvalidJSONError as e:
            return False

    def _full_path(self, path=''):
        if path:
            return 'notification/{}'.format(path)
        return 'notification'



from ..client import ResponseStructureError


STATE_ENABLED = 'OK'
STATE_DISABLED = 'ERROR'


class HealthManager:
    def __init__(self, client):
        self._client = client

    def get_notifier_state(self):
        """
        Returns current Moira Notifier state
        :return: str

        :raises: ResponseStructureError
        """
        result = self._client.get(self._full_path("notifier"))
        if 'state' not in result:
            raise ResponseStructureError("state doesn't exist in response", result)

        return result['state']

    def disable_notifications(self):
        """
        Manage Moira Notifier to stop sending notifications
        Returns current Moira Notifier state
        :return: str

        :raises: ResponseStructureError
        """
        data = {
            'state': STATE_DISABLED
        }
        result = self._client.put(self._full_path("notifier"), json=data)
        if 'state' not in result:
            raise ResponseStructureError("state doesn't exist in response", result)

        return result['state']

    def enable_notifications(self):
        """
        Manage Moira Notifier to start sending notifications
        Returns current Moira Notifier state
        :return: str

        :raises: ResponseStructureError
        """
        data = {
            'state': STATE_ENABLED
        }
        result = self._client.put(self._full_path("notifier"), json=data)
        if 'state' not in result:
            raise ResponseStructureError("state doesn't exist in response", result)

        return result['state']

    def _full_path(self, path=''):
        if path:
            return 'health/{}'.format(path)
        return 'health'



from ..client import ResponseStructureError


class WebContact:
    def __init__(self, _type, label, validation=None, placeholder=None, _help=None):
        self.type = _type
        self.label = label
        self.validation = validation
        self.placeholder = placeholder
        self.help = _help


class Config:
    def __init__(self, remote_allowed, contacts, support_email=None):
        self.remoteAllowed = remote_allowed
        self.contacts = contacts
        self.supportEmail = support_email


class ConfigManager:
    def __init__(self, client):
        self._client = client

    def fetch(self):
        """
        Returns config, see https://moira.readthedocs.io/en/latest/installation/configuration.html

        :return: config

        :raises: ResponseStructureError
        """
        result = self._client.get(self._full_path())

        if 'contacts' not in result:
            raise ResponseStructureError("'contacts' field doesn't exist in response", result)
        if 'remoteAllowed' not in result:
            raise ResponseStructureError("'remoteAllowed' field doesn't exist in response", result)

        contacts = []
        for contact in result['contacts']:
            if self._validate_contact(contact):
                _help = contact['help'] if 'help' in contact else None
                validation = contact['validation'] if 'validation' in contact else None
                placeholder = contact['placeholder'] if 'placeholder' in contact else None
                contacts.append(WebContact(_type=contact['type'], label=contact['label'], validation=validation,
                                           placeholder=placeholder, _help=_help))
        support = result['supportEmail'] if 'supportEmail' in result else None
        return Config(remote_allowed=result['remoteAllowed'], contacts=contacts, support_email=support)

    def _validate_contact(self, contact):
        if 'type' not in contact:
            return False
        if 'label' not in contact:
            return False
        return True

    def _full_path(self, path=''):
        if path:
            return 'config/{}'.format(path)
        return 'config'



from .client import Client
from .models.contact import ContactManager
from .models.event import EventManager
from .models.notification import NotificationManager
from .models.pattern import PatternManager
from .models.subscription import SubscriptionManager
from .models.tag import TagManager
from .models.trigger import TriggerManager
from .models.health import HealthManager
from .models.config import ConfigManager
from .models.user import UserManager


class Moira:
    def __init__(self, api_url, auth_custom=None,
                 auth_user=None, auth_pass=None, login=None):
        """
        :param api_url: str API URL
        :param auth_custom: dict auth custom headers
        :param auth_user: str auth user
        :param auth_pass: str auth password
        :param login: str auth login
        """
        self._client = Client(api_url, auth_custom,
                              auth_user, auth_pass, login)

        self._trigger = None
        self._tag = None
        self._event = None
        self._notification = None
        self._contact = None
        self._pattern = None
        self._subscription = None
        self._health = None
        self._config = None
        self._user = None

    @property
    def trigger(self):
        """
        Get trigger manager

        :return: TriggerManager
        """
        if not self._trigger:
            self._trigger = TriggerManager(self._client)
        return self._trigger

    @property
    def tag(self):
        """
        Get tag manager

        :return: TagManager
        """
        if not self._tag:
            self._tag = TagManager(self._client)
        return self._tag

    @property
    def event(self):
        """
        Get event manager

        :return: EventManager
        """
        if not self._event:
            self._event = EventManager(self._client)
        return self._event

    @property
    def notification(self):
        """
        Get notification manager

        :return: NotificationManager
        """
        if not self._notification:
            self._notification = NotificationManager(self._client)
        return self._notification

    @property
    def contact(self):
        """
        Get contact manager

        :return: ContactManager
        """
        if not self._contact:
            self._contact = ContactManager(self._client)
        return self._contact

    @property
    def pattern(self):
        """
        Get pattern manager

        :return: PatternManager
        """
        if not self._pattern:
            self._pattern = PatternManager(self._client)
        return self._pattern

    @property
    def subscription(self):
        """
        Get subscription manager

        :return: SubscriptionManager
        """
        if not self._subscription:
            self._subscription = SubscriptionManager(self._client)
        return self._subscription

    @property
    def health(self):
        """
        Get health manager

        :return: HealthManager
        """
        if not self._health:
            self._health = HealthManager(self._client)
        return self._health

    @property
    def config(self):
        """
        Get config manager

        :return: ConfigManager
        """
        if not self._config:
            self._config = ConfigManager(self._client)
        return self._config

    @property
    def user(self):
        """
        Get user manager

        :return: UserManager

        """
        if not self._user:
            self._user = UserManager(self._client)
        return self._user



DAYS_OF_WEEK = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']

MINUTES_IN_HOUR = 60


def get_schedule(start_hour, start_minute, end_hour, end_minute, disabled_days):
    days = []
    for day in DAYS_OF_WEEK:
        day_info = {
            'enabled': True if day not in disabled_days else False,
            'name': day
        }
        days.append(day_info)
    return {
        'days': days,
        'startOffset': start_hour * MINUTES_IN_HOUR + start_minute,
        'endOffset': end_hour * MINUTES_IN_HOUR + end_minute,
    }



from collections import namedtuple

from ..client import InvalidJSONError
from ..client import ResponseStructureError
from .trigger import Trigger


Pattern = namedtuple('Pattern', ['metrics', 'pattern', 'triggers'])


class PatternManager:
    """
    A Graphite pattern is a single dot-separated metric name, possibly containing one or more wildcards.
    """
    def __init__(self, client):
        self._client = client

    def fetch_all(self):
        """
        Returns all existing patterns in all triggers

        :return: list of Pattern

        :raises: ResponseStructureError
        """
        result = self._client.get(self._full_path())
        if 'list' in result:
            patterns = []
            for pattern in result['list']:
                if 'triggers' in pattern:
                    pattern['triggers'] = [Trigger(self._client, **trigger) for trigger in pattern['triggers']]
                patterns.append(Pattern(**pattern))
            return patterns
        else:
            raise ResponseStructureError("list doesn't exist in response", result)

    def delete(self, pattern):
        """
        Delete pattern
        Returns True even if pattern doesn't exist

        :param pattern: str pattern
        :return: True if deleted, False otherwise

        :raises: ResponseStructureError
        """
        try:
            self._client.delete(self._full_path(pattern))
            return False
        except InvalidJSONError:
            return True

    def _full_path(self, path=''):
        if path:
            return 'pattern/{}'.format(path)
        return 'pattern'



from ..client import ResponseStructureError
from ..client import InvalidJSONError
from .base import Base
from .common import MINUTES_IN_HOUR, get_schedule

STATE_OK = 'OK'
STATE_WARN = 'WARN'
STATE_ERROR = 'ERROR'
STATE_NODATA = 'NODATA'
STATE_EXCEPTION = 'EXCEPTION'

RISING_TRIGGER = 'rising'
FALLING_TRIGGER = 'falling'
EXPRESSION_TRIGGER = 'expression'


class Trigger(Base):

    def __init__(
            self,
            client,
            name,
            tags,
            targets,
            warn_value=None,
            error_value=None,
            desc='',
            ttl=600,
            ttl_state=STATE_NODATA,
            sched=None,
            expression='',
            trigger_type=None,
            is_remote=False,
            mute_new_metrics=False,
            alone_metrics=None,
            **kwargs):
        """

        :param client: api client
        :param name: str trigger name
        :param tags: list of str tags for trigger
        :param targets: list of str targets
        :param warn_value: float warning value (if T1 <= warn_value)
        :param error_value: float error value (if T1 <= error_value)
        :param desc: str trigger description
        :param ttl: int set ttl_state if has no value for ttl seconds
        :param ttl_state: str state after ttl seconds without data (one of STATE_* constants)
        :param sched: dict schedule for trigger
        :param expression: str c-like expression
        :param trigger_type: str trigger type
        :param is_remote: bool use remote storage
        :param mute_new_metrics: bool mute new metrics
        :param alone_metrics: dict with targets of alone metrics
        :param kwargs: additional parameters
        """
        self._client = client

        self._id = kwargs.get('id', None)
        self.name = name
        self.tags = tags
        self.targets = targets
        self.warn_value = warn_value
        self.error_value = error_value
        self.desc = desc
        self.ttl = ttl
        self.ttl_state = ttl_state
        
        default_sched = {
            'startOffset': 0,
            'endOffset': 1439,
            'tzOffset': 0
        }
        if not sched:
            sched = default_sched
            self.disabled_days = set()
        else:
            if 'days' in sched:
                self.disabled_days = {day['name'] for day in sched['days'] if not day['enabled']}
        self.sched = sched
        self.expression = expression
        self.trigger_type = self.resolve_type(trigger_type)

        # compute time range
        self._start_hour = self.sched['startOffset'] // MINUTES_IN_HOUR
        self._start_minute = self.sched['startOffset'] - self._start_hour * MINUTES_IN_HOUR
        self._end_hour = self.sched['endOffset'] // MINUTES_IN_HOUR
        self._end_minute = self.sched['endOffset'] - self._end_hour * MINUTES_IN_HOUR

        self.is_remote = is_remote
        self.mute_new_metrics = mute_new_metrics
        self.alone_metrics = alone_metrics

    def resolve_type(self, trigger_type):
        """
        Resolve type of a trigger
        :return: str
        """
        if trigger_type in (RISING_TRIGGER, FALLING_TRIGGER, EXPRESSION_TRIGGER):
            return trigger_type
        if self.expression != '':
            return EXPRESSION_TRIGGER
        if self.warn_value is not None and self.error_value is not None:
            if self.warn_value > self.error_value:
                return FALLING_TRIGGER
            if self.warn_value < self.error_value:
                return RISING_TRIGGER

    def add_target(self, target):
        """
        Add pattern name

        :param target: str target pattern
        :return: None
        """
        self.targets.append(target)

    def add_tag(self, tag):
        """
        Add tag to trigger

        :param tag: str tag name
        :return: None
        """
        self.tags.append(tag)

    def disable_day(self, day):
        """
        Disable day

        :param day: str one of DAYS_OF_WEEK
        :return: None
        """
        self.disabled_days.add(day)

    def enable_day(self, day):
        """
        Enable day

        :param day: str one of DAYS_OF_WEEK
        :return: None
        """
        self.disabled_days.remove(day)

    @property
    def id(self):
        return self._id

    def _send_request(self, trigger_id=None):
        data = {
            'name': self.name,
            'tags': self.tags,
            'targets': self.targets,
            'warn_value': self.warn_value,
            'error_value': self.error_value,
            'desc': self.desc,
            'ttl': self.ttl,
            'ttl_state': self.ttl_state,
            'sched': self.sched,
            'expression': self.expression,
            'is_remote': self.is_remote,
            'trigger_type': self.trigger_type,
            'mute_new_metrics': self.mute_new_metrics,
            'alone_metrics': self.alone_metrics,
        }

        if trigger_id:
            data['id'] = trigger_id
            api_response = TriggerManager(
                self._client).fetch_by_id(trigger_id)

        data['sched'] = get_schedule(self._start_hour, self._start_minute, self._end_hour, self._end_minute,
                                     self.disabled_days)

        if trigger_id and api_response:
            res = self._client.put('trigger/{id}'.format(id=trigger_id), json=data)
        else:
            res = self._client.put('trigger', json=data)
        if 'id' not in res:
            raise ResponseStructureError('id not in response', res)
        self._id = res['id']
        return self._id

    def save(self):
        """
        Save trigger

        :return: trigger_id
        """
        if self._id:
            return self.update()
        trigger = self.check_exists()

        if trigger:
            self._id = trigger.id
            self.update()
            return trigger.id

        return self._send_request()

    def update(self):
        """
        Update trigger

        :return: trigger id
        """
        return self._send_request(self._id)

    def set_start_hour(self, hour):
        """
        Set start hour

        :param hour: int hour
        :return: None
        """
        self._start_hour = int(hour)

    def set_start_minute(self, minute):
        """
        Set start minute

        :param minute: int minute
        :return: None
        """
        self._start_minute = int(minute)

    def set_end_hour(self, hour):
        """
        Set end hour

        :param hour: int hour
        :return: None
        """
        self._end_hour = int(hour)

    def set_end_minute(self, minute):
        """
        Set end minute

        :param minute: int minute
        :return: None
        """
        self._end_minute = int(minute)

    def check_exists(self):
        """
        Check if current trigger exists

        :return: trigger id if exists, None otherwise
        """
        trigger_manager = TriggerManager(self._client)
        for trigger in trigger_manager.fetch_all():
            if self.name == trigger.name and \
                set(self.targets) == set(trigger.targets) and \
                    set(self.tags) == set(trigger.tags):
                return trigger

    def get_metrics(self, start, end):
        """
        Get metrics associated with certain trigger

        :param start: The start period of metrics to get. Example : -1hour
        :param end: The end period of metrics to get. Example : now

        :return: Metrics for trigger
        """
        try:
            params = {
                'from': start,
                'to': end,
            }
            result = self._client.get('trigger/{id}/metrics'.format(id=self.id), params=params)
            return result
        except InvalidJSONError:
            return []

    def delete_metric(self, metric_name):
        """
        Deletes metric from last check and all trigger pattern metrics

        :param metric_name: str name of the target metric, example: DevOps.my_server.hdd.freespace_mbytes

        :return: True if success, False otherwise
        """
        try:
            params = {
                'name': metric_name,
            }
            self._client.delete('trigger/{id}/metrics'.format(id=self.id), params=params)
            return True
        except InvalidJSONError:
            return False

    def delete_nodata_metrics(self):
        """
        Deletes all metrics from last data which are in NODATA state.
        It also deletes all trigger patterns of those metrics.

        :return: True if success, False otherwise
        """
        try:
            self._client.delete('trigger/{id}/metrics/nodata'.format(id=self.id))
            return True
        except InvalidJSONError:
            return False


class TriggerManager:
    def __init__(self, client):
        self._client = client

    @property
    def trigger_client(self):
        return self._client

    def fetch_all(self):
        """
        Returns all existing triggers

        :return: list of Trigger

        :raises: ResponseStructureError
        """
        result = self._client.get(self._full_path())
        if 'list' in result:
            triggers = []
            for trigger in result['list']:
                triggers.append(Trigger(self._client, **trigger))
            return triggers
        else:
            raise ResponseStructureError("list doesn't exist in response", result)

    def fetch_by_id(self, trigger_id):
        """
        Returns Trigger by trigger id

        :param trigger_id: str trigger id
        :return: Trigger

        :raises: ResponseStructureError
        """
        result = self._client.get(self._full_path('{id}/state'.format(id=trigger_id)))
        if 'state' in result:
            trigger = self._client.get(self._full_path(trigger_id))
            return Trigger(self._client, **trigger)
        elif not 'trigger_id' in result:
            raise ResponseStructureError("invalid api response", result)

    def search(self, only_problems, page, text):
        """
        Search triggers

        :param only_problems: Restricts the result to errors only. Example: false
        :param page: Defines the number of the displayed page. E.g, page=2 would display the 2nd page. Example: 1
        :param text: Query to perform a search for. Example: cpu

        :return: matching triggers list

        :raises: ResponseStructureError
        """
        params = {
            'onlyProblems': only_problems,
            'page': page,
            'text': text,
        }
        result = self._client.get(self._full_path(), params=params)
        if 'list' not in result:
            raise ResponseStructureError("list doesn't exist in response", result)

        return result['list']

    def delete(self, trigger_id):
        """
        Delete trigger by trigger id

        :param trigger_id: str trigger id
        :return: True if deleted, False otherwise
        """
        try:
            self._client.delete(self._full_path(trigger_id))
            return False
        except InvalidJSONError:
            return True

    def get_throttling(self, trigger_id):
        """
        Get a trigger with its throttling i.e its next allowed message time

        :param trigger_id: str trigger id
        :return: trigger throttle value or None
        """
        try:
            result = self._client.get(self._full_path('{id}/throttling'.format(id=trigger_id)))
            if 'throttling' in result:
                return result['throttling']
            return None
        except InvalidJSONError:
            return None

    def reset_throttling(self, trigger_id):
        """
        Resets throttling by trigger id

        :param trigger_id: str trigger id
        :return: True if reset, False otherwise
        """
        try:
            self._client.delete(self._full_path('{id}/throttling'.format(id=trigger_id)))
            return True
        except InvalidJSONError:
            return False

    def get_state(self, trigger_id):
        """
        Get state of trigger by trigger id

        :param trigger_id: str trigger id
        :return: state of trigger
        """
        return self._client.get(self._full_path('{id}/state'.format(id=trigger_id)))

    def get_metrics(self, trigger_id, _from, to):
        """
        Get metrics associated with certain trigger

        :param trigger_id: str trigger id
        :param _from: The start period of metrics to get. Example : -1hour
        :param to: The end period of metrics to get. Example : now

        :return: Metrics for trigger
        """
        try:
            params = {
                'from': _from,
                'to': to,
            }
            result = self._client.get(self._full_path('{id}/metrics'.format(id=trigger_id)), params=params)
            return result
        except InvalidJSONError:
            return []

    def remove_metric(self, trigger_id, metric):
        """
        Remove metric by trigger id

        :param trigger_id: str trigger id
        :param metric: str metric name
        :return: True if removed, False otherwise
        """
        try:
            params = {
                'name': metric
            }
            self._client.delete(self._full_path('{id}/metrics'.format(id=trigger_id)), params=params)
            return True
        except InvalidJSONError:
            return False

    def remove_nodata_metrics(self, trigger_id):
        """
        Remove metric by trigger id

        :param trigger_id: str trigger id
        :return: True if removed, False otherwise
        """
        try:
            self._client.delete(self._full_path('{id}/metrics/nodata'.format(id=trigger_id)))
            return True
        except InvalidJSONError:
            return False

    def is_exist(self, trigger):
        """
        Check whether trigger exists or not

        :param trigger: Trigger trigger to check
        :return: bool
        """
        for moira_trigger in self.fetch_all():
            if trigger.name == moira_trigger.name and \
                set(trigger.targets) == set(moira_trigger.targets) and \
                    set(trigger.tags) == set(moira_trigger.tags):
                return True
        return False

    def get_non_existent(self, triggers):
        """
        Returns triggers which are not exist yet

        :param triggers: list of Trigger
        :return: list of Trigger
        """
        moira_triggers = self.fetch_all()
        non_existent = []
        for trigger in triggers:
            exist = False
            for moira_trigger in moira_triggers:
                if trigger.name == moira_trigger.name and \
                                set(trigger.targets) == set(moira_trigger.targets) and \
                                set(trigger.tags) == set(moira_trigger.tags):
                    exist = True
                    break
            if not exist:
                non_existent.append(trigger)

        return non_existent

    def set_maintenance(self, trigger_id, end_time, metrics=None):
        """
        Set maintenance trigger id or metric name

        :param trigger_id: str trigger id
        :param metrics: "metric name" - "end-time" map, end-time like param end_time
        :param end_time: unix time to end the scheduled maintenance
        :return: True if success, False otherwise
        """
        try:
            data = {
                'trigger': end_time,
            }
            if metrics is not None:
                data['metrics'] = metrics
            self._client.put(self._full_path('{id}/setMaintenance'.format(id=trigger_id)), json=data)
            return True
        except InvalidJSONError:
            return False

    def create(
            self,
            name,
            tags,
            targets,
            warn_value=None,
            error_value=None,
            desc='',
            ttl=600,
            ttl_state=STATE_NODATA,
            sched=None,
            expression='',
            trigger_type=None,
            is_remote=False,
            mute_new_metrics=False,
            alone_metrics=None,
            **kwargs
    ):
        """
        Creates new trigger. To save it call save() method of Trigger.
        :param name: str trigger name
        :param tags: list of str tags for trigger
        :param targets: list of str targets
        :param warn_value: float warning value (if T1 <= warn_value)
        :param error_value: float error value (if T1 <= error_value)
        :param desc: str trigger description
        :param ttl: int set ttl_state if has no value for ttl seconds
        :param ttl_state: str state after ttl seconds without data (one of STATE_* constants)
        :param sched: dict schedule for trigger
        :param expression: str c-like expression
        :param trigger_type: str trigger type
        :param is_remote: bool use remote storage
        :param mute_new_metrics: bool mute new metrics
        :param alone_metrics: dict with targets of alone metrics
        :param kwargs: additional trigger params
        :return: Trigger
        """
        return Trigger(
            self._client,
            name,
            tags,
            targets,
            warn_value,
            error_value,
            desc,
            ttl,
            ttl_state,
            sched,
            expression,
            trigger_type,
            is_remote,
            mute_new_metrics,
            alone_metrics,
            **kwargs
        )

    def _full_path(self, path=''):
        if path:
            return 'trigger/{}'.format(path)
        return 'trigger'



from ..client import InvalidJSONError
from ..client import ResponseStructureError
from .base import Base
from .common import MINUTES_IN_HOUR, get_schedule


class Subscription(Base):
    def __init__(self, client, tags, contacts=None, enabled=None, throttling=None, sched=None,
                 ignore_warnings=False, ignore_recoverings=False, plotting=None, any_tags=False, **kwargs):
        """

        :param client: api client
        :param tags: list of str tags
        :param contacts: list of contact id's
        :param enabled: bool is enabled
        :param throttling: bool throttling
        :param sched: dict schedule
        :param ignore_warnings: bool ignore warnings
        :param ignore_recoverings: bool ignore recoverings
        :param plotting: dict plotting settings
        :param any_tags: bool any tags
        :param kwargs: additional parameters
        """
        self._client = client

        self._id = kwargs.get('id', None)
        self.tags = tags if not any_tags else []
        if not contacts:
            contacts = []
        self.contacts = contacts
        self.enabled = enabled
        self.any_tags = any_tags
        self.throttling = throttling
        default_sched = {
            'startOffset': 0,
            'endOffset': 1439,
            'tzOffset': 0
        }
        if not sched:
            sched = default_sched
            self.disabled_days = set()
        else:
            if 'days' in sched and sched['days'] is not None:
                self.disabled_days = {day['name'] for day in sched['days'] if not day['enabled']}
        self.sched = sched

        # compute time range
        self._start_hour = self.sched['startOffset'] // MINUTES_IN_HOUR
        self._start_minute = self.sched['startOffset'] - self._start_hour * MINUTES_IN_HOUR
        self._end_hour = self.sched['endOffset'] // MINUTES_IN_HOUR
        self._end_minute = self.sched['endOffset'] - self._end_hour * MINUTES_IN_HOUR

        self.ignore_warnings = ignore_warnings
        self.ignore_recoverings = ignore_recoverings

        if not plotting:
            plotting = {'enabled': False, 'theme': 'light'}
        self.plotting = plotting

    def _send_request(self, subscription_id=None):
        data = {
            'contacts': self.contacts,
            'tags': self.tags,
            'enabled': self.enabled,
            'any_tags': self.any_tags,
            'throttling': self.throttling,
            'sched': self.sched,
            'ignore_warnings': self.ignore_warnings,
            'ignore_recoverings': self.ignore_recoverings,
            'plotting': self.plotting
        }

        if subscription_id:
            data['id'] = subscription_id

        data['sched'] = get_schedule(self._start_hour, self._start_minute, self._end_hour, self._end_minute,
                                     self.disabled_days)

        if subscription_id:
            result = self._client.put('subscription/{id}'.format(id=subscription_id), json=data)
        else:
            result = self._client.put('subscription', json=data)
        if 'id' not in result:
            raise ResponseStructureError("id doesn't exist in response", result)

        self._id = result['id']
        return self._id

    def disable_day(self, day):
        """
        Disable day

        :param day: str one of DAYS_OF_WEEK
        :return: None
        """
        self.disabled_days.add(day)

    def enable_day(self, day):
        """
        Enable day

        :param day: str one of DAYS_OF_WEEK
        :return: None
        """
        self.disabled_days.remove(day)

    def add_tag(self, tag):
        """
        Add tag to subscription

        :param tag: str tag name
        :return: None
        """
        self.tags.append(tag)

    def add_contact(self, contact_id):
        """
        Add contact

        :param contact_id: str contact id
        :return: None
        """
        self.contacts.append(contact_id)

    def enable_plotting(self, theme='light'):
        """
        Enable plotting

        :param theme: str plotting theme
        :return: None
        """
        self.plotting = {
            'enabled': True,
            'theme': theme
            }

    def disable_plotting(self):
        """
        Disable plotting

        :return: None
        """
        self.plotting = {
            'enabled': False,
            'theme': 'light'
            }

    def save(self):
        """
        Save subscription

        :return: subscription id
        """
        if self._id:
            return self.update()
        self._send_request()

    def update(self):
        """
        Update subscription

        :return: subscription id
        """
        if not self._id:
            return self.save()
        self._send_request(self._id)

    def set_start_hour(self, hour):
        """
        Set start hour

        :param hour: int hour

        :return: None
        """
        self._start_hour = hour

    def set_start_minute(self, minute):
        """
        Set start minute

        :param minute: int minute

        :return: None
        """
        self._start_minute = minute

    def set_end_hour(self, hour):
        """
        Set end hour

        :param hour: int hour

        :return: None
        """
        self._end_hour = hour

    def set_end_minute(self, minute):
        """
        Set end minute

        :param minute: int minute

        :return: None
        """
        self._end_minute = minute


class SubscriptionManager:
    def __init__(self, client):
        self._client = client

    def fetch_all(self):
        """
        Returns all existing subscriptions

        :return: list of Subscription

        :raises: ResponseStructureError
        """
        result = self._client.get(self._full_path())
        if 'list' in result:
            subscriptions = []
            for subscription in result['list']:
                subscriptions.append(Subscription(self._client, **subscription))
            return subscriptions
        else:
            raise ResponseStructureError("list doesn't exist in response", result)

    def is_exist(self, **kwargs):
        """
        Check whether subscription exists or not by any attributes

        :param kwargs: attributes
        :return: bool

        :raises: ValueError
        """
        for subscription in self.fetch_all():
            equal = True
            for attr, value in kwargs.items():
                try:
                    if getattr(subscription, attr) != value:
                        equal = False
                        break
                except Exception:
                    raise ValueError('Wrong attribute "{}"'.format(attr))
            if equal:
                return True
        return False

    def create(self, tags, contacts=None, enabled=True, throttling=True, sched=None,
               ignore_warnings=False, ignore_recoverings=False, plotting=None, any_tags=False, **kwargs):
        """
        Create new subscription.
        
        :param tags: list of str tags
        :param contacts: list of contact id's
        :param enabled: bool is enabled
        :param throttling: bool throttling
        :param sched: dict schedule
        :param ignore_warnings: bool ignore warnings
        :param ignore_recoverings: bool ignore recoverings
        :param kwargs: additional parameters
        :param plotting: dict plotting settings
        :param any_tags: bool any tags
        :return: Subscription
        """

        return Subscription(
            self._client,
            tags,
            contacts, 
            enabled,
            throttling,
            sched,
            ignore_warnings,
            ignore_recoverings,
            plotting,
            any_tags,
            **kwargs
        )

    def delete(self, subscription_id):
        """
        Remove subscription by given id

        :return: True on success, False otherwise
        """
        try:
            self._client.delete(self._full_path(subscription_id))
            return False
        except InvalidJSONError as e:
            if e.content == b'':  # successfully if response is blank
                return True
            return False

    def test(self, subscription_id):
        """
        Send test notification to subscription contact

        :return: True on success, False otherwise
        """
        try:
            self._client.put(self._full_path('{id}/test'.format(id=subscription_id)))
            return False
        except InvalidJSONError as e:
            if e.content == b'':  # successfully if response is blank
                return True
            return False

    def _full_path(self, path=''):
        if path:
            return 'subscription/{}'.format(path)
        return 'subscription'



from ..client import InvalidJSONError
from ..client import ResponseStructureError


MAX_FETCH_LIMIT = 1000


class EventManager:
    def __init__(self, client):
        self._client = client

    def fetch_by_trigger(self, trigger, limit=MAX_FETCH_LIMIT):
        """
        Get all events by trigger
        :param trigger: Trigger trigger
        :param limit: int limit
        :return: list of dicts

        :raises: ValueError
        :raises: ResponseStructureError
        """
        if not trigger.id:
            raise ValueError('Trigger id is None')
        params = {
            'p': 0,
            'size': limit
        }
        result = self._client.get(self._full_path(trigger.id), params=params)
        if 'list' not in result:
            raise ResponseStructureError("list doesn't exist in response", result)

        return result['list']

    def delete_all(self):
        """
        Remove all events

        :return: True on success, False otherwise
        """
        try:
            result = self._client.delete(self._full_path("all"))
            return False
        except InvalidJSONError as e:
            if e.content == b'':  # successfully if response is blank
                return True
            return False

    def _full_path(self, path=''):
        if path:
            return 'event/{}'.format(path)
        return 'event'



from requests.auth import HTTPBasicAuth
import requests


class ResponseStructureError(Exception):
    def __init__(self, msg, content):
        """

        :param msg: str error message
        :param content: dict response content
        """
        self.msg = msg
        self.content = content


class InvalidJSONError(Exception):
    def __init__(self, content):
        """

        :param content: bytes response content
        """
        self.content = content


class Client:
    def __init__(self, api_url, auth_custom=None, auth_user=None, auth_pass=None, login=None):
        """

        :param api_url: str Moira API URL
        :param auth_custom: dict auth custom headers
        :param auth_user: str auth user
        :param auth_pass: str auth password
        :param login: str auth login
        """
        if not api_url.endswith('/'):
            self.api_url = api_url + '/'
        else:
            self.api_url = api_url

        if not login and auth_user:
            login = auth_user

        self.auth = None
        self.headers = {
            'X-Webauth-User': login,
            'Content-Type': 'application/json',
            'User-Agent': 'Python Moira Client'
            }

        if auth_user and auth_pass:
            self.auth = HTTPBasicAuth(auth_user, auth_pass)

        if auth_custom:
            self.headers.update(auth_custom)

    def get(self, path='', **kwargs):
        """

        :param path: str api path
        :param kwargs: additional parameters for request
        :return: dict response

        :raises: HTTPError
        :raises: InvalidJSONError
        """
        r = requests.get(self._path_join(path), headers=self.headers, auth=self.auth, **kwargs)
        r.raise_for_status()
        try:
            return r.json()
        except ValueError:
            raise InvalidJSONError(r.content)

    def delete(self, path='', **kwargs):
        """

        :param path: str api path
        :param kwargs: additional parameters for request
        :return: dict response

        :raises: HTTPError
        :raises: InvalidJSONError
        """
        r = requests.delete(self._path_join(path), headers=self.headers, auth=self.auth, **kwargs)
        r.raise_for_status()
        try:
            return r.json()
        except ValueError:
            raise InvalidJSONError(r.content)

    def put(self, path='', **kwargs):
        """

        :param path: str api path
        :param kwargs: additional parameters for request
        :return: dict response

        :raises: HTTPError
        :raises: InvalidJSONError
        """
        r = requests.put(self._path_join(path), headers=self.headers, auth=self.auth, **kwargs)
        r.raise_for_status()
        try:
            return r.json()
        except ValueError:
            raise InvalidJSONError(r.content)

    def post(self, path='', **kwargs):
        """

        :param path: str api path
        :param kwargs: additional parameters for request
        :return: dict response

        :raises: HTTPError
        :raises: InvalidJSONError
        """
        r = requests.post(self._path_join(path), headers=self.headers, auth=self.auth, **kwargs)
        r.raise_for_status()
        try:
            return r.json()
        except ValueError:
            raise InvalidJSONError(r.content)

    def _path_join(self, *args):
        path = self.api_url
        for part in args:
            path += part
        return path



from collections import namedtuple

from requests.exceptions import HTTPError

from ..client import InvalidJSONError
from ..client import ResponseStructureError
from .subscription import Subscription


TagStats = namedtuple('TagStats', ['name', 'subscriptions', 'triggers'])


class TagManager:
    def __init__(self, client):
        self._client = client

    def fetch_all(self):
        """
        Returns all existing tags

        :return: list of str

        :raises: ResponseStructureError
        """
        result = self._client.get(self._full_path())
        if 'list' not in result:
            raise ResponseStructureError("list doesn't exist in response", result)

        return result['list']

    def delete(self, tag):
        """
        Delete tag.
        In case if tag doesn't exist returns True.
        Returns False if tag is assigned to at least 1 trigger.

        :param tag: str tag name
        :return: True if deleted, False otherwise
        """
        try:
            self._client.delete(self._full_path(tag))
        except (InvalidJSONError, HTTPError):
            return False

        return True

    def stats(self):
        """
        Returns stats by all triggers

        :return: list of TagStats

        :raises: ResponseStructureError
        """
        result = self._client.get(self._full_path('stats'))
        if 'list' in result:
            for stat in result['list']:
                if 'subscriptions' in stat:
                    stat['subscriptions'] = [
                        Subscription(self._client, **subscription) for subscription in stat['subscriptions']
                        ]
            return [TagStats(**stat) for stat in result['list']]
        else:
            raise ResponseStructureError("list doesn't exist in response", result)

    def fetch_assigned_triggers(self, tag):
        """
        Returns triggers assigned to tag

        :param tag: str tag name
        :return: list of trigger id's

        :raises: ResponseStructureError
        """
        result = self._client.get(self._full_path('stats'))
        if 'list' in result:
            trigger_ids = set()
            for stat in result['list']:
                if 'triggers' in stat and 'name' in stat:
                    if stat['name'] == tag:
                        for trigger_id in stat['triggers']:
                            trigger_ids.add(trigger_id)
                        return list(trigger_ids)
            return list(trigger_ids)
        else:
            raise ResponseStructureError("list doesn't exist in response", result)

    def fetch_assigned_triggers_by_tags(self, tags):
        """
        Returns triggers assigned to at least one tag of tags

        :param tags: Iterable of tags
        :return: list of trigger id's

        :raises: ResponseStructureError
        """

        tags = set(tags)
        result = self._client.get(self._full_path('stats'))
        if 'list' in result:
            trigger_ids = set()
            for stat in result['list']:
                if 'triggers' in stat and 'name' in stat:
                    if stat['name'] in tags:
                        for trigger_id in stat['triggers']:
                            trigger_ids.add(trigger_id)

            return list(trigger_ids)
        else:
            raise ResponseStructureError("list doesn't exist in response", result)

    def fetch_assigned_subscriptions(self, tag):
        """
        Returns subscriptions assigned to tag

        :param tag: str tag name
        :return: list of str subscription_id

        :raises: ResponseStructureError
        """
        result = self._client.get(self._full_path('stats'))
        if 'list' in result:
            trigger_ids = set() 
            for stat in result['list']:
                if 'subscriptions' in stat and 'name' in stat:
                    if stat['name'] == tag:
                        return [
                            Subscription(self._client, **subscription) for subscription in stat['subscriptions']
                            ]
            return list(trigger_ids)
        else:
            raise ResponseStructureError("list doesn't exist in response", result)

    def _full_path(self, path=''):
        if path:
            return 'tag/{}'.format(path)
        return 'tag'




class Base:
    @property
    def id(self):
        return self._id

    def __repr__(self):
        return '({} {})'.format(self.__class__.__name__, self.id)

    def __unicode__(self):
        return u'({} {})'.format(self.__class__.__name__, self.id)

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id



from distutils.core import setup

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='moira-python-client',
    version='2.6.0',
    description='Client for Moira - Alerting system based on Graphite data',
    keywords='moira monitoring client metrics alerting',
    long_description="""
        Moira is a real-time alerting tool, based on Graphite data.
        moira-client is a python client for Moira API.
        Key features:
        - create, update, delete, manage triggers
        - create, delete, update subscriptions
        - manage tags, patterns, notifications, events, contacts
    """,
    author = 'Alexander Lukyanchenko',
    author_email = 'al.lukyanchenko@gmail.com',
    packages=[
        'moira_client',
        'moira_client.models'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',

        'Operating System :: OS Independent',
        'Intended Audience :: Developers',

        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
        'Topic :: System :: Monitoring',

        "License :: OSI Approved :: MIT License"
    ],
    url='https://github.com/moira-alert/python-moira-client',
    install_requires=required
)



from ..client import InvalidJSONError
from ..client import ResponseStructureError
from .base import Base

CONTACT_EMAIL = 'mail'
CONTACT_PUSHOVER = 'pushover'
CONTACT_SLACK = 'slack'
CONTACT_TELEGRAM = 'telegram'
CONTACT_TWILIO_SMS = 'twilio sms'
CONTACT_TWILIO_VOICE = 'twilio voice'


class Contact(Base):
    def __init__(self, value='', type='', **kwargs):
        """

        :param value: str contact value
        :param contact_type: str contact type (one of CONTACT_* constants)
        :param kwargs: additional parameters
        """
        self.type = type
        self.value = value
        self.user = kwargs.get('user', None)
        self._id = kwargs.get('id', None)


class ContactManager:
    def __init__(self, client):
        self._client = client

    def add(self, value, contact_type):
        """
        Add new contact

        :param value: str contact value
        :param contact_type: str contact type (one of CONTACT_* constants)
        :return: Contact

        :raises: ResponseStructureError
        """
        data = {
            'value': value,
            'type': contact_type
        }

        contacts = self.fetch_by_current_user()
        for contact in contacts:
            if contact.value == value and contact.type == contact_type:
                return contact

        result = self._client.put(self._full_path(), json=data)
        if 'id' not in result:
            raise ResponseStructureError('No id in response', result)

        return Contact(id=result['id'], **data)

    def fetch_all(self):
        """
        Returns all existing contacts

        :return: list of Contact

        :raises: ResponseStructureError
        """
        result = self._client.get(self._full_path())
        if 'list' not in result:
            raise ResponseStructureError("list doesn't exist in response", result)

        contacts = []

        for contact in result['list']:
            contacts.append(Contact(**contact))

        return contacts

    def fetch_by_current_user(self):
        """
        Returns all contacts by current user

        :return: list of Contact

        :raises: ResponseStructurteError
        """
        result = self._client.get('user/settings')
        if 'contacts' not in result:
            raise ResponseStructureError("'contacts' field doesn't exist in response", result)

        contacts = []

        for contact in result['contacts']:
            contacts.append(Contact(**contact))

        return contacts

    def get_id(self, type, value):
        """
        Returns contact id by type and value
        Returns None if contact doesn't exist

        :param type: str contact type
        :param value: str contact value
        :return: str contact id
        """
        for contact in self.fetch_all():
            if contact.type == type and contact.value == value:
                return contact.id

    def delete(self, contact_id):
        """
        Delete contact by contact id
        If contact id doesn't exist returns True

        :param contact_id: str contact id
        :return: True if ok, False otherwise

        :raises: ResponseStructureError
        """
        try:
            self._client.delete(self._full_path(contact_id))
            return False
        except InvalidJSONError as e:
            if e.content == b'':  # successfully if response is blank
                return True
            else:
                return False

    def update(self, contact):
        """
        Updates an existing notification contact

        :param contact: Contact
        :return: True if ok, False otherwise
        """
        data = {
            'type': contact.type,
            'value': contact.value,
        }

        try:
            self._client.put(self._full_path(contact.id), json=data)
            return True
        except InvalidJSONError:
            return False

    def test(self, contact_id):
        """
        Push a test notification to verify that the contact is properly set up.

        :param contact_id: str contact id
        :return: True if ok, False otherwise
        """

        try:
            self._client.post(self._full_path('{id}/test'.format(id=contact_id)))
            return False
        except InvalidJSONError as e:
            if len(e.content) == 0:  # successfully if response is blank
                return True
            else:
                return False

    def _full_path(self, path=''):
        if path:
            return 'contact/{}'.format(path)
        return 'contact'
